from typing import List
from tooldelta.launch_cli import FrameFBConn, FrameNeOmg, FrameNeOmgRemote
from tooldelta.cfg import Cfg

PRG_NAME = "ToolDelta"

PLUGIN_MARKET_SOURCE_OFFICIAL = "https://mirror.ghproxy.com/raw.githubusercontent.com/ToolDelta/ToolDelta-PluginMarket/main"

LAUNCHERS: List[
    tuple[str, type[FrameFBConn | FrameNeOmg | FrameNeOmgRemote]]
] = [
    (
        "FastBuilder External 模式 (经典模式) §c(已停止维护, 无法适应新版本租赁服!)",
        FrameFBConn,
    ),
    ("NeOmega 框架 (NeOmega模式, 租赁服适应性强, 推荐)", FrameNeOmg),
    (
        "NeOmega 框架 (NeOmega连接模式, 需要先启动对应的neOmega接入点)",
        FrameNeOmgRemote,
    ),
]

LAUNCH_CFG: dict = {
    "服务器号": 0,
    "密码": 0,
    "启动器启动模式(请不要手动更改此项, 改为0可重置)": 0,
    "验证服务器地址(更换时记得更改fbtoken)": "",
    "是否记录日志": True,
    "插件市场源": PLUGIN_MARKET_SOURCE_OFFICIAL
}

LAUNCH_CFG_STD: dict = {
    "服务器号": int,
    "密码": int,
    "启动器启动模式(请不要手动更改此项, 改为0可重置)": Cfg.NNInt,
    "验证服务器地址(更换时记得更改fbtoken)": str,
    "是否记录日志": bool,
    "插件市场源": str
}

FB_APIS = [
    "https://api.fastbuilder.pro/api/phoenix/login",
    "https://api.fastbuilder.pro/api/new",
    "https://api.fastbuilder.pro/api/api",
    "https://api.fastbuilder.pro/api/login",
    "https://api.fastbuilder.pro",
]


GUGU_APIS = [
    "https://liliya233.uk/api/phoenix/login",
    "https://liliya233.uk/api/new",
    "https://liliya233.uk",
]

AUTH_SERVERS = [
    ("FastBuilder 官方验证服务器", "https://api.fastbuilder.pro"),
    ("咕咕酱 FB验证服务器", "https://liliya233.uk"),
]

TOOLDELTA_PLUGIN_DIR = "插件文件"
TOOLDELTA_CLASSIC_PLUGIN = "ToolDelta类式插件"
TOOLDELTA_INJECTED_PLUGIN = "ToolDelta注入式插件"



import time
import ujson as json
from typing import TYPE_CHECKING, Any
from typing import Optional
from tooldelta.color_print import Print
from tooldelta.packets import Packet_CommandOutput
if TYPE_CHECKING:
    from tooldelta.frame import Frame, GameCtrl


def check_avaliable(sth: "GameCtrl") -> Optional[AttributeError]:
    """
    检查给定的 "GameCtrl" 对象是否可用

    参数:
        sth: 要检查的 "GameCtrl" 对象
    """
    if sth is None:
        raise AttributeError(f"无法使用 {sth.__class__.__name__}, 因为其还未被初始化")


game_control: "GameCtrl"
movent_frame: "Frame"


def set_frame(my_Frame: "Frame") -> None:
    """
    全局初始化框架

    参数:
        my_"Frame": 要设置的框架对象
    """
    global movent_frame, game_control  # pylint: disable=global-statement
    movent_frame = my_Frame
    game_control = my_Frame.get_game_control()


def sendcmd(
    cmd: str, waitForResp: bool = False, timeout: int = 30
) -> None | Packet_CommandOutput:
    """发送命令到游戏控制器，并可选择是否等待响应\n
    如果 waitForResp 为 False,则返回 None,否则返回 Packet_CommandOutput 对象

    参数:
        cmd: 要发送的命令
        waitForResp: 是否等待响应,默认为 False
        timeout: 等待响应的超时时间（秒）,默认为 30
    """
    check_avaliable(game_control)
    if game_control.sendcmd is None:
        raise AttributeError(f"无法使用 {game_control.__class__.__name__}, 因为其还未被初始化")
    return game_control.sendcmd(cmd, waitForResp, timeout)


def sendwscmd(
    cmd: str, waitForResp: bool = False, timeout: int | float = 30
) -> Packet_CommandOutput:
    """
    发送WSCMD命令到游戏控制器

    参数:
        cmd: 要发送的WSCMD命令
        waitForResp: 是否等待响应 默认为False
        timeout: 超时时间（秒） 默认为30
    """
    check_avaliable(game_control)
    return game_control.sendwscmd(cmd, waitForResp, timeout)


def sendwocmd(cmd: str) -> None:
    """
    发送WO命令到游戏控制器

    参数:
        cmd: 要发送的WO命令
    """
    check_avaliable(game_control)
    game_control.sendwocmd(cmd)


def sendPacket(pktID: int, pkt: str) -> None:
    """
    发送数据包给游戏控制器

    参数:
        pktID: 数据包ID
        pkt: 数据包内容
    """
    check_avaliable(game_control)
    game_control.sendPacket(pktID, pkt)


def sendfbcmd(cmd: str) -> None:
    """向FastBuilder发送命令\n
    在除FastBuilder外的其他启动器上不可用

    参数:
        cmd: 要发送的命令。
    """
    check_avaliable(game_control)
    game_control.sendfbcmd(cmd)


def rawText(playername: str, text: str) -> None:
    """
    向指定玩家发送原始文本消息

    参数:
        playername: 玩家名称
        text: 要发送的文本
    """
    sendcmd(r"""/tellraw %s {"rawtext":[{"text":"%s"}]}""" % (playername, text))


def tellrawText(playername: str, title: str | None = None, text: str = "") -> None:
    """
    向指定玩家发送tellraw消息

    参数:
        playername: 玩家名称
        title: 标题文本（可选）
        text: 消息文本
    """
    if title is None:
        sendcmd(r"""/tellraw %s {"rawtext":[{"text":"§r%s"}]}""" % (playername, text))
    else:
        sendcmd(
            r"""/tellraw %s {"rawtext":[{"text":"<%s> §r%s"}]}"""
            % (
                playername,
                title,
                text,
            )
        )


def get_all_player() -> list:
    """获取所有玩家列表"""
    check_avaliable(game_control)
    return game_control.allplayers


def is_op(playername: str) -> bool | None:
    """
    判断玩家是否为OP

    参数:
        playername: 玩家名称
    """
    check_avaliable(game_control)
    if playername not in get_all_player():
        return False
    # 检测框架是否为"Frame"NeOmg
    if movent_frame.launcher.is_op is not None:
        return movent_frame.launcher.is_op(playername)


def getTarget(sth: str, timeout: bool | int = 5) -> list:
    """
    获取符合目标选择器实体的列表

    参数:
        sth: 目标选择器
        timeout: 超时时间，默认为5秒
    """
    check_avaliable(game_control)
    if not sth.startswith("@"):
        raise Exception("Minecraft Target Selector is not correct.")
    result = (
        game_control.sendwscmd(f"/testfor {sth}", True, timeout)
        .OutputMessages[0]
        .Parameters
    )
    if result:
        result = result[0]
        return result.split(", ")
    return []


def find_key_from_value(dic: dict, val: Any) -> Optional[Any]:
    """
    从字典中根据值查找对应的键

    参数:
        dic: 目标字典
        val: 目标值
    """
    for k, v in dic.items():
        if v == val:
            return k


def get_robotname() -> str | None:
    """获取机器人名称。"""
    check_avaliable(game_control)
    return game_control.bot_name


def getPos(targetNameToGet: str, timeout: float | int = 5) -> dict:
    """获取目标玩家的位置信息

    参数:
        targetNameToGet: 目标玩家的名称
        timeout: 超时时间（秒）。默认为5秒

    异常:
        ValueError: 当目标玩家不存在时抛出该异常
        ValueError: 当获取位置信息失败时抛出该异常
        AttributeError: 当获取玩家UUID失败时抛出该异常
    """
    check_avaliable(game_control)
    if targetNameToGet not in get_all_player() or targetNameToGet.startswith("@"):
        raise ValueError(f"Player {targetNameToGet} does not exist.")
    result = sendwscmd(f'/querytarget @a[name="{targetNameToGet}"]', True, timeout)
    if not result.OutputMessages[0].Success:
        raise ValueError(f"Failed to get the position: {result.OutputMessages[0]}")
    parameter = result.OutputMessages[0].Parameters[0]
    if isinstance(parameter, str):
        resultList = json.loads(parameter)
    else:
        resultList = parameter
    result = {}

    if game_control.players_uuid is None:
        raise AttributeError("Failed to get the players_uuid.")
    targetName = targetNameToGet
    x = (
        parameter[0]["position"]["x"]
        if parameter[0]["position"]["x"] >= 0
        else parameter[0]["position"]["x"] - 1
    )
    y = parameter[0]["position"]["y"] - 1.6200103759765
    z = (
        parameter[0]["position"]["z"]
        if parameter[0]["position"]["z"] >= 0
        else parameter[0]["position"]["z"] - 1
    )
    position = {
        "x": float(f"{x:.2f}"),
        "y": float(f"{y:.2f}"),
        "z": float(f"{z:.2f}"),
    }
    dimension = parameter[0]["dimension"]
    yRot = parameter[0]["yRot"]
    result[targetName] = {
        "dimension": dimension,
        "position": position,
        "yRot": yRot,
    }
    if targetNameToGet == "@a":
        return result
    if len(result) != 1:
        raise ValueError("Failed to get the position.")
    if targetNameToGet.startswith("@a"):
        return list(result.values())[0]
    return result[targetNameToGet]


def countdown(delay: int | float, msg: str | None = None) -> None:
    """
    倒计时函数

    参数:
        delay: 延迟时间，可以是整数或浮点数
        msg: 倒计时消息，可选参数，默认为"Countdown"
    """
    if msg is None:
        msg = "Countdown"
    delayStart = time.time()
    delayStop = delayStart + delay
    while delay >= 0:
        if delay >= 0:
            delay = delayStop - time.time()
        else:
            delay = 0
        Print.print_inf(f"{msg}: {delay:.2f}s", end="\r")
        time.sleep(0.01)
    print("", end="\n")


def getBlockTile(x: int, y: int, z: int) -> str:
    """
    获取指定坐标的方块的ID

    参数:
        x: X坐标
        y: Y坐标
        z: Z坐标
    """
    res = sendwscmd(f"/testforblock {x} {y} {z} air", True)
    if res.SuccessCount:
        return "air"
    return res.OutputMessages[0].Parameters[4].strip("%tile.").strip(".name")


def getTickingAreaList() -> dict | AttributeError:
    """
    获取 tickingarea 列表

    异常:
        AttributeError: 获取 tickingarea 列表失败
    """
    result = {}
    cmd = sendwscmd("/tickingarea list all-dimensions", True)
    if cmd is not None:
        resultList = cmd.OutputMessages
    else:
        raise AttributeError("Failed to get the tickingarea list.")

    if resultList[0].Success is False:
        return result
    for tickareaData in resultList[1].Message.split("%dimension.dimensionName")[1:]:
        tickareaDimension = tickareaData.split(": \n")[0]
        tickareaList = tickareaData.split(": \n")[1].split("\n")
        for tickarea in tickareaList:
            if not tickarea:
                continue
            tickareaName = tickarea.split("- ", 1)[1].split(": ", 1)[0]
            tickareaPos = {
                "start": tickarea.split(" ")[2:5],
                "end": tickarea.split(" ")[6:9],
            }
            tickareaPos["start"].pop(1)
            tickareaPos["end"].pop(1)
            result[tickareaName] = {"dimension": tickareaDimension}
            result[tickareaName].update(tickareaPos)
    return result



import asyncio
import importlib
import threading
import time
import traceback
from typing import Any, Callable, Type

from tooldelta.color_print import Print
from tooldelta.plugin_load.classic_plugin import Plugin
from tooldelta.plugin_load import (
    classic_plugin,
    injected_plugin,
    NON_FUNC,
    NotValidPluginError,
    PluginAPINotFoundError,
    PluginAPIVersionError,
)


class PluginGroup:
    plugins: list[Plugin] = []
    plugins_funcs: dict[str, list] = {
        "on_def": [],
        "on_inject": [],
        "on_player_prejoin": [],
        "on_player_join": [],
        "on_player_message": [],
        "on_player_death": [],
        "on_player_leave": [],
    }
    plugin_added_cache = {"plugin": None, "packets": []}
    pluginAPI_added_cache = []
    _broadcast_evts_cache = {}

    def __init__(self):
        self._listen_packet_ids = set()
        self._packet_funcs: dict[str, list[Callable]] = {}
        self._plugins_api: dict[str, Plugin] = {}
        self._broadcast_listeners: dict[str, list[Callable]] = {}
        self.excType = 0
        self.normal_plugin_loaded_num = 0
        self.injected_plugin_loaded_num = 0
        self.loaded_plugins_name = []
        self.linked_frame = None

    def add_plugin(self, plugin: Plugin):
        try:
            if not Plugin.__subclasscheck__(plugin):
                raise NotValidPluginError(f"插件主类必须继承Plugin类 而不是 {plugin}")
        except TypeError:
            if not Plugin.__subclasscheck__(type(plugin)):
                raise NotValidPluginError(
                    f"插件主类必须继承Plugin类 而不是 {plugin.__class__.__name__}"
                )
        self.plugin_added_cache["plugin"] = plugin
        self.test_plugin(plugin)
        return plugin

    def add_plugin_as_api(self, apiName: str):
        def _add_plugin_2_api(api_plugin: Type[Plugin]):
            if not Plugin.__subclasscheck__(api_plugin):
                raise NotValidPluginError("API插件主类必须继承Plugin类")
            self.plugin_added_cache["plugin"] = api_plugin
            self.pluginAPI_added_cache.append(apiName)
            self.test_plugin(api_plugin)
            return api_plugin

        return _add_plugin_2_api

    def add_packet_listener(self, pktID: int | list[int]):
        """
        将下面的方法作为一个MC数据包接收器
        装饰器参数:
            pktID: int | list[int], 数据包ID或多个ID
        接收器方法参数:
            pkt: dict, 数据包包体
        """
        def deco(func):
            if isinstance(pktID, int):
                self.plugin_added_cache["packets"].append((pktID, func))
            else:
                for i in pktID:
                    self.plugin_added_cache["packets"].append((i, func))
            return func

        return deco

    def add_broadcast_listener(self, evt_name: str):
        """
        将下面的方法作为一个广播事件接收器
        装饰器参数:
            evt_name: str, 事件类型
        传入方法接受参数:
            data: Any, 若事件带有附加参数, 则会传入附加参数, 否则传入None
        返回参数: Any(返回的数据, 会被广播事件源收集, 默认None)
        原理:
        方法1 广播: hi, what's ur name? 附加参数=english_only
            方法2 接收到广播并被执行: 方法2(english_only) -> my name is Super. -> 收集表
        事件1 获取到 收集表 作为返回: ["my name is Super."]
        """

        def deco(func: Callable[[Any], bool]):
            if self._broadcast_evts_cache.get(evt_name):
                self._broadcast_evts_cache[evt_name].append(func)
            else:
                self._broadcast_evts_cache[evt_name] = [func]
            return func

        return deco

    def broadcastEvt(self, evt_name: str, data: Any = None) -> list[Any] | None:
        """
        向全局广播一个特定事件, 可以传入附加信息参数
        参数:
            evt_name: str, 事件名
            **kwargs: 附加信息参数
        返回:
            收集到的数据的列表(如果接收到广播的方法返回了数据的话)
        """
        callback_list = []
        res = self._broadcast_listeners.get(evt_name)
        if res:
            for f in res:
                res2 = f(data)
                if res2:
                    callback_list.append(res2)
        return callback_list

    def test_plugin(self, plugin: Plugin):
        if self.linked_frame is None:
            # 很可能是直接单独运行此插件的代码.
            Print.clean_print(f"插件主类信息({plugin.name}): ")
            Print.clean_print(f" - 作者: {plugin.author}\n - 版本: {plugin.version}")
            Print.clean_print(
                f" - 数据包监听: {', '.join(str(i) for i in self._listen_packet_ids)}"
            )
    @staticmethod
    def help(plugin: Plugin):
        """
        查看插件帮助.
        常用于查看 get_plugin_api() 方法获取到的插件实例的帮助.
        """
        plugin_docs = "<plugins.help>: " + plugin.name + "开放的API接口说明:\n"
        for attr_name, attr in plugin.__dict__.items():
            if not attr_name.startswith("__") and attr.__doc__ is not None:
                plugin_docs += (
                    "\n §a"
                    + attr_name
                    + ":§f\n    "
                    + attr.__doc__.replace("\n", "\n    ")
                )
        Print.clean_print(plugin_docs)

    def checkSystemVersion(self, need_vers: tuple[int, int, int]):
        """检查ToolDelta系统的最低版本"""
        if need_vers > self.linked_frame.sys_data.system_version:
            raise self.linked_frame.SystemVersionException(
                f"该组件需要{self.linked_frame.PRG_NAME}为{'.'.join([str(i) for i in self.linked_frame.sys_data.system_version])}版本"
            )

    def get_plugin_api(self, apiName: str, min_version: tuple | None = None) -> Plugin:
        """
        通过插件API名获取插件实例
        参数:
            apiName: 插件API名
            min_version: API最低版本(若不填则默认不检查最低版本)
        """
        api = self._plugins_api.get(apiName)
        if api:
            if min_version and api.version < min_version:
                raise PluginAPIVersionError(apiName, min_version, api.version)
            return api
        raise PluginAPINotFoundError(f"无法找到API插件：{apiName}")

    def set_frame(self, frame):
        self.linked_frame = frame

    def read_all_plugins(self):
        try:
            classic_plugin.read_plugins(self)
            self.execute_def(self.linked_frame.on_plugin_err)
            asyncio.run(injected_plugin.load_plugin(self))
        except Exception:
            err_str = "\n".join(traceback.format_exc().split("\n")[1:])
            Print.print_err(f"加载插件出现问题: \n{err_str}")
            raise SystemExit

    @staticmethod
    def load_plugin_hot(plugin_name: str, plugin_type: str):
        plugin = None
        if plugin_type == "classic":
            plugin = classic_plugin.load_plugin(plugin_name, True)
        elif plugin_type == "injected":
            asyncio.run(injected_plugin.load_plugin_file(plugin_name))
        if plugin is not None:
            plugin.on_def()
        Print.print_suc(f"成功热加载插件: {plugin_name}")

    def _add_listen_packet_id(self, packetType: int):
        self._listen_packet_ids.add(packetType)
        self.linked_frame.link_game_ctrl.add_listen_pkt(packetType)

    def _add_listen_packet_func(self, packetType, func: Callable):
        if self._packet_funcs.get(str(packetType)):
            self._packet_funcs[str(packetType)].append(func)
        else:
            self._packet_funcs[str(packetType)] = [func]

    def _add_broadcast_evt(self, evt, func):
        if self._broadcast_listeners.get(evt):
            self._broadcast_listeners[evt].append(func)
        else:
            self._broadcast_listeners[evt] = [func]

    def execute_def(self, onerr: Callable[[str, Exception, str], None] = NON_FUNC):
        for name, func in self.plugins_funcs["on_def"]:
            try:
                func()
            except PluginAPINotFoundError as err:
                Print.print_err(f"插件 {name} 需要包含该种接口的前置组件: {err.name}")
                raise SystemExit
            except PluginAPIVersionError as err:
                Print.print_err(
                    f"插件 {name} 需要该前置组件 {err.name} 版本: {err.m_ver}, 但是现有版本过低: {err.n_ver}"
                )
                raise SystemExit
            except Exception as err:
                onerr(name, err, traceback.format_exc())

    def execute_init(self, onerr: Callable[[str, Exception, str], None] = NON_FUNC):
        for name, func in self.plugins_funcs["on_inject"]:
            try:
                func()
            except Exception as err:
                onerr(name, err, traceback.format_exc())

    def execute_player_prejoin(
        self, player, onerr: Callable[[str, Exception, str], None] = NON_FUNC
    ):
        for name, func in self.plugins_funcs["on_player_prejoin"]:
            try:
                func(player)
            except Exception as err:
                onerr(name, err, traceback.format_exc())

    def execute_player_join(
        self, player, onerr: Callable[[str, Exception, str], None] = NON_FUNC
    ):
        for name, func in self.plugins_funcs["on_player_join"]:
            try:
                func(player)
            except Exception as err:
                onerr(name, err, traceback.format_exc())

    def execute_player_message(
        self, player, msg, onerr: Callable[[str, Exception, str], None] = NON_FUNC
    ):
        pat = f"[{player}] "
        if msg.startswith(pat):
            msg = msg.strip(pat)
        for name, func in self.plugins_funcs["on_player_message"]:
            try:
                func(player, msg)
            except Exception as err:
                onerr(name, err, traceback.format_exc())

    def execute_player_leave(
        self, player, onerr: Callable[[str, Exception, str], None] = NON_FUNC
    ):
        for name, func in self.plugins_funcs["on_player_leave"]:
            try:
                func(player)
            except Exception as err:
                onerr(name, err, traceback.format_exc())

    def execute_player_death(
        self,
        player: str,
        killer: str | None,
        msg: str,
        onerr: Callable[[str, Exception, str], None] = NON_FUNC,
    ):
        for name, func in self.plugins_funcs["on_player_death"]:
            try:
                func(player, killer, msg)
            except Exception as err:
                onerr(name, err, traceback.format_exc())

    def processPacketFunc(self, pktID: int, pkt: dict):
        d = self._packet_funcs.get(str(pktID))
        if d:
            for func in d:
                try:
                    res = func(pkt)
                    if res:
                        return True
                except:
                    Print.print_err(f"插件方法 {func.__name__} 出错：")
                    Print.print_err(traceback.format_exc())
        return False


plugin_group = PluginGroup()



import time
import os
import re
import sys
import threading
import json
import traceback
import datetime
import platform
import socketio
import subprocess
import socket
import logging
import ctypes
import asyncio
import multiprocessing
import copy
import math
import random
import psutil
import pyspeedtest
import requests
import nbt
import qrcode
import getpass
import pymysql
import websockets
import tarfile
import gzip
import importlib
import ujson
import hashlib
import base64
import rich
import tqdm
import psutil
from typing import Callable
sys.path.append(os.path.join(os.getcwd(), "libs"))



from .basic_mods import *
from .color_print import Print
from .packets import Packet_CommandOutput
from .plugin_load.PluginGroup import Plugin
from .starter import plugin_group as plugins
from .starter import start_tool_delta, safe_jump
from .frame import Config, Builtins, Frame, GameCtrl
from .launch_options import client_title


import importlib
import os
import sys
import traceback
import typing
from tooldelta.color_print import Print
from tooldelta.builtins import Builtins
from tooldelta.cfg import Cfg
from tooldelta.plugin_load.funcs import unzip_plugin
from tooldelta.plugin_load import plugin_is_enabled, NotValidPluginError
from tooldelta.constants import TOOLDELTA_CLASSIC_PLUGIN

if typing.TYPE_CHECKING:
    # define Frame
    from tooldelta.frame import Frame
    from tooldelta.plugin_load.PluginGroup import PluginGroup


class Plugin:
    name: str = ""
    version = (0, 0, 1)
    author = "?"
    description = "..."

    def __init__(self, frame: "Frame"):
        self.frame = frame
        self.game_ctrl = frame.get_game_control()

    @property
    def data_path(self) -> str:
        return os.path.join("插件数据文件", self.name)


plugin_group_cache:list["PluginGroup"] = [None]


def read_plugins(plugin_grp: "PluginGroup"):
    plugin_group_cache[0] = plugin_grp
    PLUGIN_PATH = os.path.join("插件文件", TOOLDELTA_CLASSIC_PLUGIN)
    sys.path.append(os.path.join("插件文件", TOOLDELTA_CLASSIC_PLUGIN))
    for plugin_dir in os.listdir(PLUGIN_PATH):
        if not plugin_is_enabled(plugin_dir):
            continue
        if (
            not os.path.isdir(os.path.join(PLUGIN_PATH, plugin_dir.strip(".zip")))
            and os.path.isfile(os.path.join(PLUGIN_PATH, plugin_dir))
            and plugin_dir.endswith(".zip")
        ):
            Print.print_with_info(f"§6正在解压插件{plugin_dir}, 请稍后", "§6 解压 ")
            unzip_plugin(
                os.path.join(PLUGIN_PATH, plugin_dir),
                os.path.join(PLUGIN_PATH, plugin_dir.strip(".zip")),
            )
            Print.print_suc(f"§a成功解压插件{plugin_dir} -> 插件目录")
            plugin_dir = plugin_dir.strip(".zip")
        if os.path.isdir(os.path.join(PLUGIN_PATH, plugin_dir)):
            sys.path.append(os.path.join(PLUGIN_PATH, plugin_dir))
            load_plugin(plugin_dir)
            plugin_grp.loaded_plugins_name.append(plugin_dir)


def load_plugin(plugin_dirname: str, hot_load=False):
    plugin_grp = plugin_group_cache[0]
    try:
        if os.path.isfile(
            os.path.join(
                "插件文件", TOOLDELTA_CLASSIC_PLUGIN, plugin_dirname, "__init__.py"
            )
        ):
            importlib.__import__(plugin_dirname)
        else:
            Print.print_war(f"{plugin_dirname} 文件夹 未发现插件文件, 跳过加载")
            return
        Builtins.simpleAssert(
            plugin_grp.plugin_added_cache["plugin"] is not None,
            NotValidPluginError(
                "需要调用1次 @plugins.add_plugin 以注册插件主类, 然而没有调用"
            ),
        )
        plugin: type[Plugin] = plugin_grp.plugin_added_cache["plugin"]
        if plugin.name is None:
            raise ValueError(f"插件主类 {plugin.__name__} 需要作者名")
        plugin_ins = plugin(plugin_grp.linked_frame)
        plugin_grp.plugins.append([plugin_ins.name, plugin_ins])
        _v0, _v1, _v2 = plugin_ins.version
        for evt_name in (
            "on_def",
            "on_inject",
            "on_player_prejoin",
            "on_player_join",
            "on_player_message",
            "on_player_death",
            "on_player_leave"
        ):
            if hasattr(plugin_ins, evt_name):
                plugin_grp.plugins_funcs[evt_name].append(
                    [plugin_ins.name, getattr(plugin_ins, evt_name)]
                )
        Print.print_suc(
            f"成功载入插件 {plugin_ins.name} 版本: {_v0}.{_v1}.{_v2} 作者：{plugin_ins.author}"
        )
        plugin_grp.normal_plugin_loaded_num += 1
        if plugin_grp.plugin_added_cache["packets"] != []:
            for pktType, func in plugin_grp.plugin_added_cache["packets"]:  # type: ignore
                pfunc = getattr(plugin_ins, func.__name__)
                if pfunc is None:
                    raise NotValidPluginError("数据包监听不能在主插件类以外定义")
                plugin_grp._add_listen_packet_id(pktType)
                plugin_grp._add_listen_packet_func(
                    pktType, pfunc
                )
        if plugin_grp.pluginAPI_added_cache is not None:
            for _api in plugin_grp.pluginAPI_added_cache:
                if isinstance(_api, str):
                    plugin_grp._plugins_api[_api] = plugin_ins
                else:
                    (apiName, api) = _api
                    plugin_grp._plugins_api[apiName] = api(plugin_grp.linked_frame)
        if plugin_grp._broadcast_evts_cache != {}:
            for evt, funcs in plugin_grp._broadcast_evts_cache.items():
                for func in funcs:
                    bfunc = getattr(plugin_ins, func.__name__)
                    if bfunc is not None:
                        # 在插件主类以内定义了广播接收器
                        func = bfunc
                    plugin_grp._add_broadcast_evt(evt, func)
        return plugin_ins
    except NotValidPluginError as err:
        Print.print_err(f"插件 {plugin_dirname} 不合法: {err.args[0]}")
        raise SystemExit
    except Cfg.ConfigError as err:
        Print.print_err(f"插件 {plugin_dirname} 配置文件报错：{err}")
        Print.print_err("你也可以直接删除配置文件, 重新启动ToolDelta以自动生成配置文件")
        raise SystemExit
    except Builtins.SimpleJsonDataReader.DataReadError as err:
        Print.print_err(f"插件 {plugin_dirname} 读取数据失败: {err}")
    except plugin_grp.linked_frame.SystemVersionException:
        Print.print_err(f"插件 {plugin_dirname} 需要更高版本的ToolDelta加载: {err}")
    except Exception as err:
        Print.print_err(f"加载插件 {plugin_dirname} 出现问题, 报错如下: ")
        Print.print_err("§c" + traceback.format_exc())
        raise SystemExit
    finally:
        plugin_grp.plugin_added_cache["plugin"] = None
        plugin_grp.plugin_added_cache["packets"].clear()
        plugin_grp.pluginAPI_added_cache.clear()
        plugin_grp._broadcast_evts_cache.clear()
    return None



import os
import re
import shutil
import socket
import platform
import time
from typing import Union
from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm import tqdm
import pyspeedtest
import requests

from tooldelta.color_print import Print

def progress_bar(
    current: float | int, total: float | int, length = 20, color1="§f", color2="§b"
):
    pc = round(current / total * length)
    return Print.colormode_replace(
        color1 + " " * pc + color2 + " " * (20 - pc) + "§r ", 7
    )

def download_progress_bar(
    current_bytes: int, total_bytes: int, speed = 0
):
    bar = progress_bar(current_bytes, total_bytes)
    b = f"{bar} {_pretty_kb(current_bytes)}B / {_pretty_kb(total_bytes)}B"
    if speed != 0:
        b += f"{_pretty_kb(speed)}B/s)    "
    Print.print_with_info(b, "§a 下载 ", need_log=False, end = "\r")

def _pretty_kb(n: int) -> str:
    if n >= 1048576:
        return f"{round(n / 1048576, 2)}M"
    if n >= 1024:
        return f"{round(n / 1024, 2)}K"
    return f"{round(n, 1)}"


def _path_get_filename(path: str) -> Union[str, None]:
    "从路径中获取最末尾的文件夹/文件名."
    if "/" not in path:
        return None
    return path.split("/")[-1]


def _is_common_text_file(url_path: str) -> bool:
    "是否为正常的文本文件(常见的)."
    return any(
        url_path.endswith(i)
        for i in [".txt", ".yml", ".md", ".xml", ".py", ".h", ".c", ".pyi", ".json"]
    )


def get_file_size(url: str) -> Union[int, None]:
    response = requests.head(url, timeout=10)
    if "Content-Length" in response.headers:
        file_size = int(response.headers["Content-Length"])
        return file_size
    return None


def download_file_chunk(url: str, start_byte: int, end_byte: int, save_dir: str) -> int:
    headers = {"Range": f"bytes={start_byte}-{end_byte}"}
    response = requests.get(url, headers=headers, stream=True, timeout=10)
    response.raise_for_status()
    with open(save_dir + ".tmp", "rb+") as dwnf:
        dwnf.seek(start_byte)
        downloaded_bytes = 0
        for chunk in response.iter_content(chunk_size=8192):
            dwnf.write(chunk)
            downloaded_bytes += len(chunk)
        return downloaded_bytes


def download_file_singlethreaded(
    url: str, save_dir: str, ignore_warnings: bool = False
) -> None:
    with requests.get(url, stream=True, timeout=10) as res:
        res.raise_for_status()
        filesize = get_file_size(url)
        try:
            if filesize < 256 and not ignore_warnings:
                Print.print_war(f"下载 {url} 的文件警告: 文件大小异常, 不到 0.25KB")
        except TypeError:
            pass
        chunk_size = 8192
        nowsize = 0
        succ = False
        lastime = time.time()
        useSpeed = 0
        with open(save_dir + ".tmp", "wb") as dwnf:
            for chk in res.iter_content(chunk_size):
                nowtime = time.time()
                if nowtime != lastime:
                    useSpeed = chunk_size / (nowtime - lastime)
                    lastime = nowtime
                nowsize += len(chk)
                dwnf.write(chk)
                if nowsize % 81920 == 0:  # 每下载 10 个数据块更新一次进度
                    download_progress_bar(nowsize, filesize, useSpeed)
        succ = True
        if succ:
            shutil.move(save_dir + ".tmp", save_dir)
        else:
            os.remove(save_dir + ".tmp")

def download_file_multithreading(
    url: str, save_dir: str, num_threads: int = 8, ignore_warnings: bool = False
):
    filesize = get_file_size(url)
    if filesize is None:
        download_file_singlethreaded(url=url, save_dir=save_dir)
    if filesize < 256 and not ignore_warnings:
        Print.print_war(f"下载 {url} 的文件警告: 文件大小异常, 不到 0.25KB")
    elif filesize < 1048576 and not ignore_warnings:
        download_file_singlethreaded(url=url, save_dir=save_dir)
    chunk_size = filesize // num_threads  # 每个线程下载的块大小
    with open(save_dir + ".tmp", "wb") as dwnf, Print.lock, tqdm(
        total=filesize,
        unit="B",
        unit_scale=True,
        desc=Print.fmt_info("", "§a 下载 §r"),
        ncols=80,
    ) as pbar:

        def update_progress(downloaded_bytes: int) -> None:
            pbar.update(downloaded_bytes)

        with ThreadPoolExecutor(max_workers=num_threads) as executor:
            futures = []
            for i in range(num_threads):
                start_byte = i * chunk_size
                end_byte = start_byte + chunk_size - 1
                if i == num_threads - 1:
                    end_byte = filesize - 1
                future = executor.submit(
                    download_file_chunk, url, start_byte, end_byte, save_dir
                )
                future.add_done_callback(lambda f: update_progress(f.result()))
                futures.append(future)
            for future in futures:
                future.result()
    shutil.move(save_dir + ".tmp", save_dir)


def download_unknown_file(url: str, save_dir: str) -> None:
    # 鉴于 Content-Length 不一定表示文件原始大小, 二进制文件与文本文件需要分开下载
    # 否则显示的下载条会异常
    resp = requests.get(url, timeout=10)
    resp.raise_for_status()
    # Bad..
    if _is_common_text_file(save_dir):
        # 文本文件, 体积可能不大
        with open(save_dir, "wb") as f:
            f.write(resp.content)
    else:
        download_file_singlethreaded(url, save_dir)

def test_site_latency(Da: dict) -> list:
    tmp_speed = {}
    urls = [Da["url"]] + Da["mirror_url"]

    with ThreadPoolExecutor() as executor:
        futures = [executor.submit(measure_latencyt, url) for url in urls]

        for future, url in zip(as_completed(futures), urls):
            try:
                latency = future.result(timeout=5)
                if latency != -1:
                    tmp_speed[url] = latency
            except Exception as e:
                Print.print_war(f"Error measuring latency for {url}: {e}")

    sorted_speed = sorted(tmp_speed.items(), key=lambda x: x[1], reverse=True)
    return sorted_speed

def measure_latencyt(url: str) -> float:
    try:
        st = pyspeedtest.SpeedTest(
            re.search(r"(?<=http[s]://)[.\w-]*(:\d{,8})?((?=/)|(?!/))", url).group()
        )
        download_speed = st.download()  # / 1000000  # 转换为兆字节/秒
        return download_speed
    except Exception as e:
        Print.print_war(f"Error measuring latency for {url}: {e}")
        return -1.0  # 返回-1表示测速失败


def get_free_port(start: int = 8080, end: int = 65535) -> int:
    for port in range(start, end):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind(("localhost", port))
                return port
            except OSError:
                Print.print_war(f"端口 {port} 正被占用, 跳过")
    raise Exception(f"未找到空闲端口({start}~{end})")



import os
import platform
import shutil
import shlex
import json
from tooldelta.builtins import Builtins
from tooldelta.color_print import Print
from tooldelta.plugin_market import market
from tooldelta.plugin_load import PluginRegData
from tooldelta.constants import (
    TOOLDELTA_PLUGIN_DIR,
    TOOLDELTA_CLASSIC_PLUGIN,
    TOOLDELTA_INJECTED_PLUGIN
)

JsonIO = Builtins.SimpleJsonDataReader

if platform.system().lower() == "windows":
    CLS_CMD = "cls"
else:
    CLS_CMD = "clear"

clear_screen = lambda: os.system(shlex.quote(CLS_CMD))

class PluginManager:
    plugin_reg_data_path = "插件注册表"
    default_reg_data = {"classic": {}, "injected": {}, "unknown": {}}
    _plugin_datas_cache = []

    def manage_plugins(self):
        # 进入插件管理界面
        clear_screen()
        np = self.register_plugins_auto()
        if np > 0:
            Print.clean_print(f"§a已自动注册{np}个未被注册的插件.")
        while 1:
            plugins = self.list_plugins_list()
            Print.clean_print("§f输入§bu§f更新本地所有插件, §f输入§cq§f退出")
            r = input(Print.clean_fmt("§f输入插件关键词进行选择\n(空格可分隔关键词):"))
            if r.strip() == "":
                continue
            if r.lower() == "q":
                return
            if r.lower() == "u":
                self.update_all_plugins(
                    self.get_plugin_reg_name_dict_and_datas()[1]
                )
            else:
                res = self.search_plugin(r, plugins)
                if res is None:
                    input()
                else:
                    self.plugin_operation(res)
            clear_screen()

    def plugin_operation(self, plugin: PluginRegData):
        # 对插件进行操作
        description_fixed = plugin.description.replace('\n', '\n    ')
        clear_screen()
        Print.clean_print(f"§d插件名: §f{plugin.name}")
        Print.clean_print(f" - 版本: {plugin.version_str}")
        Print.clean_print(f" - 作者: {plugin.author}")
        Print.clean_print(f" 描述: {description_fixed}")
        Print.clean_print(f"§f1.删除插件  2.检查更新  3.{'禁用插件' if plugin.is_enabled else '启用插件'}")
        f_dirname = {
            "classic": TOOLDELTA_CLASSIC_PLUGIN,
            "injected": TOOLDELTA_INJECTED_PLUGIN
        }[plugin.plugin_type]
        match input(Print.clean_fmt("§f请选择选项: ")):
            case "1":
                r = input(
                    Print.clean_fmt("§c删除插件操作不可逆, 请输入y, 其他取消: ")
                ).lower()
                if r != "y":
                    return
                plugin_dir = os.path.join("插件文件", f_dirname, plugin.name)
                shutil.rmtree(
                    plugin_dir + ("+disabled" if not plugin.is_enabled else "")
                )
                Print.clean_print(f"§a已成功删除插件 {plugin.name}, 回车键继续")
                self.pop_plugin_reg_data(plugin)
                input()
                return
            case "2":
                latest_version = market.get_latest_plugin_version(plugin.plugin_id)
                if latest_version is None:
                    Print.clean_print("§6无法获取其的最新版本, 回车键继续")
                elif latest_version == plugin.version_str:
                    Print.clean_print("§a此插件已经为最新版本, 回车键继续")
                else:
                    Print.clean_print(f"§a插件有新版本可用 ({plugin.version_str} => {latest_version})")
                    r = input(Print.clean_fmt("输入§a1§f=立刻更新, §62§f=取消更新: ")).strip()
                    if r == "1":
                        Print.clean_print("§a正在下载新版插件...", end = "\r")
                        market.download_plugin(plugin, market.get_datas_from_market())
                        Print.clean_print("§a插件更新完成, 回车键继续")
                        plugin.version = (int(i) for i in latest_version.split("."))
                    else:
                        Print.clean_print("§6已取消, 回车键返回")
            case "3":
                if plugin.is_enabled:
                    os.rename(
                        os.path.join("插件文件", f_dirname, plugin.name),
                        os.path.join("插件文件", f_dirname, plugin.name + "+disabled")
                    )
                else:
                    os.rename(
                        os.path.join("插件文件", f_dirname, plugin.name + "+disabled"),
                        os.path.join("插件文件", f_dirname, plugin.name)
                    )
                plugin.is_enabled = [True, False][plugin.is_enabled]
                Print.clean_print(f"§6当前插件状态为: {['§c禁用', '§a启用'][plugin.is_enabled]}")
        self.push_plugin_reg_data(plugin)
        input()

    def update_all_plugins(self, plugins: list[PluginRegData]):
        market_datas = self.latest_version = market.get_datas_from_market()["MarketPlugins"]
        need_updates: list[tuple[PluginRegData, str]] = []
        for i in plugins:
            s_data = market_datas.get(i.plugin_id)
            if s_data is None:
                continue
            if i.version_str != s_data["version"]:
                need_updates.append((i, s_data["version"]))
        if need_updates:
            clear_screen()
            Print.clean_print("§f以下插件可进行更新:")
            for plugin, v in need_updates:
                Print.clean_print(f" - {plugin.name} §6{plugin.version_str}§f -> §a{v}")
            r = input(Print.clean_fmt("§f输入§a y §f开始更新, §c n §f取消: ")).strip().lower()
            if r == "y":
                for plugin, v in need_updates:
                    market.download_plugin(plugin, market_datas)
                Print.clean_print("§a全部插件已更新完成")
            else:
                Print.clean_print("§6已取消插件更新.")
            input("[Enter键继续...]")
        else:
            input(Print.clean_fmt("§a无可更新的插件. [Enter键继续]"))

    def search_plugin(self, resp, plugins):
        res = self.search_plugin_by_kw(resp.split(" "), plugins)
        if res == []:
            Print.clean_print("§c没有任何已安装插件匹配得上关键词")
            return None
        if len(res) > 1:
            Print.clean_print("§a☑ §f关键词查找到的插件:")
            for i, plugin in enumerate(res):
                Print.clean_print(str(i + 1) + ". " + self.make_plugin_icon(plugin))
            r = Builtins.try_int(input(Print.clean_fmt("§f请选择序号: ")))
            if r is None or r not in range(1, len(res) + 1):
                Print.clean_print("§c序号无效, 回车键继续")
                return None
            return res[r - 1]
        return res[0]

    @staticmethod
    def search_plugin_by_kw(kws: list[str], plugins: list[PluginRegData]):
        res = []
        for plugin in plugins:
            if all(kw in plugin.name for kw in kws):
                res.append(plugin)
        return res

    def plugin_is_registered(self, plugin_name: str):
        if not self._plugin_datas_cache:
            _, self._plugin_datas_cache = self.get_plugin_reg_name_dict_and_datas()
        for i in self._plugin_datas_cache:
            if i.name == plugin_name:
                return True
        return False

    def register_plugins_auto(self):
        dirs = [TOOLDELTA_CLASSIC_PLUGIN, TOOLDELTA_INJECTED_PLUGIN]
        any_plugin_registered = 0
        for f_dir in dirs:
            dirs_type = {TOOLDELTA_CLASSIC_PLUGIN: "classic", TOOLDELTA_INJECTED_PLUGIN: "injected"}[f_dir]
            for plugin_path in os.listdir(os.path.join(TOOLDELTA_PLUGIN_DIR, f_dir)):
                datpath = os.path.join(TOOLDELTA_PLUGIN_DIR, f_dir, plugin_path, "datas.json")
                if not self.plugin_is_registered(dirs_type) and os.path.isfile(datpath):
                    with open(datpath, "r", encoding="utf-8") as f:
                        jsdata = json.load(f)
                        self.push_plugin_reg_data(PluginRegData(plugin_path, jsdata))
                        any_plugin_registered += 1
        return any_plugin_registered

    def push_plugin_reg_data(self, plugin_data: PluginRegData):
        # 向插件注册表推送插件注册信息
        r = JsonIO.readFileFrom(
            "主系统核心数据", self.plugin_reg_data_path, self.default_reg_data
        )
        r[plugin_data.plugin_type][plugin_data.name] = plugin_data.dump()
        JsonIO.writeFileTo("主系统核心数据", self.plugin_reg_data_path, r)

    def pop_plugin_reg_data(self, plugin_data: PluginRegData):
        # 从插件注册表删除插件注册信息
        r = JsonIO.readFileFrom("主系统核心数据", self.plugin_reg_data_path)
        del r[plugin_data.plugin_type][plugin_data.name]
        JsonIO.writeFileTo("主系统核心数据", self.plugin_reg_data_path, r)

    def get_plugin_reg_name_dict_and_datas(self):
        # 返回一个表示插件所在类别下的全部已注册插件的列表, 和全部已注册插件的插件注册信息列表
        r0: dict[str, list[str]] = {"classic": [], "injected": []}
        r = JsonIO.readFileFrom(
            "主系统核心数据", self.plugin_reg_data_path, self.default_reg_data
        )
        f_dirname = {
            "classic": TOOLDELTA_CLASSIC_PLUGIN,
            "injected": TOOLDELTA_INJECTED_PLUGIN
        }
        res: list[PluginRegData] = []
        for _, r1 in r.items():
            for k, v in r1.items():
                if not isinstance(k, str) or not isinstance(v, dict):
                    raise ValueError(
                        f"获取插件注册表出现问题: 类型出错: {k.__class__.__name__}, {v.__class__.__name__}"
                    )
                v.update({"name": k})
                p = PluginRegData(k, v)
                if (
                    os.path.exists(os.path.join("插件文件", f_dirname[p.plugin_type], p.name))
                    or os.path.exists(os.path.join("插件文件", f_dirname[p.plugin_type], p.name + "+disabled"))
                ):
                    res.append(p)
                    r0[p.plugin_type].append(p.name)
        return r0, res

    def get_2_compare_plugins_reg(self):
        # 返回一个全注册插件的列表
        f_plugins: list[PluginRegData] = []
        reg_dict, reg_list = self.get_plugin_reg_name_dict_and_datas()
        for p, k in {
            TOOLDELTA_CLASSIC_PLUGIN: "classic",
            TOOLDELTA_INJECTED_PLUGIN: "injected",
        }.items():
            for i in os.listdir(os.path.join("插件文件", p)):
                if i.replace("+disabled", "") not in reg_dict[k]:
                    f_plugins.append(PluginRegData(i, {"plugin-type": k}, False))
        return f_plugins + reg_list
    @staticmethod
    def make_plugin_icon(plugin: PluginRegData):
        ico_colors = { "classic": "§b", "injected": "§d"}
        return (
            ico_colors.get(plugin.plugin_type, "§7")
            + "■ "
            + (("§a"if plugin.is_enabled else "§7") if plugin.is_registered else "§6")
            + plugin.name
        )

    def make_printable_list(self, plugins: list[PluginRegData]):
        texts = []
        for plugin in plugins:
            texts.append(self.make_plugin_icon(plugin))
        lfts = []
        rgts = []
        for i, t in enumerate(texts):
            if (i + 1) % 2 == 1:
                lfts.append(t)
            else:
                rgts.append(t)
        for i, t in enumerate(lfts):
            if i in range(len(rgts)):
                Print.clean_print("§f" + Print.align(t, 35) + "§f" + Print.align(rgts[i]))
            else:
                Print.clean_print("§f" + Print.align(t, 35))

    @staticmethod
    def test_name_same(name: str, dirname: str):
        if name != dirname:
            raise AssertionError(f"插件名: {name} 与文件夹名({dirname}) 不一致") from None

    def list_plugins_list(self):
        Print.clean_print("§a☑ §f目前已安装的插件列表:")
        all_plugins = self.get_2_compare_plugins_reg()
        self.make_printable_list(all_plugins)
        return all_plugins


plugin_manager = PluginManager()



import requests, getpass, hashlib, json
from tooldelta import constants
from tooldelta.color_print import Print

def login_liliya():
    while 1:
        hash_obj: hashlib._Hash = hashlib.sha256()
        username: str = input(Print.fmt_info("请输入账号:", "§6 账号 "))
        hash_obj.update(
            getpass.getpass(Print.fmt_info("请输入密码(已隐藏):", "§6 密码 ")).encode()
        )
        __password: str = hash_obj.hexdigest()
        Authorization: str = requests.get(url=constants.GUGU_APIS[1], timeout=5).text
        repo: requests.Response = requests.post(
            url=constants.GUGU_APIS[0],
            data=json.dumps(
                {
                    "client_public_key": "",
                    "server_code": "::DRY::",
                    "server_passcode": "::DRY::",
                    "username": username,
                    "password": __password,
                },
                ensure_ascii=False,
            ),
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {Authorization}",
            }, timeout=5
        )
        repo_text: dict = json.loads(repo.text)
        repo_message: str = repo_text["message"]
        repo_success: bool = repo_text["success"]
        if repo.status_code != 200:
            Print.print_war(
                f"请求Api接口失败，将自动使用Token登陆!状态码:{repo.status_code}，返回值:{repo.text}"
            )
            return
        if not repo_success:
            if "无效的用户中心用户名或密码" in repo_message:
                Print.print_war("登录失败, 无效的用户名、密码或MFA代码!")
            else:
                Print.print_war("登录失败, 原因: " + repo_message + ", 请重试!")
        else:
            break
    token = repo_text["token"]
    with open("fbtoken", "w", encoding="utf-8") as f:
        f.write(token)

def login_fbuc():
    while 1:
        hash_obj: hashlib._Hash = hashlib.sha256()
        username: str = input(Print.fmt_info("请输入账号:", "§6 账号 "))
        hash_obj.update(
            getpass.getpass(
                Print.fmt_info("请输入密码(已隐藏):", "§6 密码 ")
            ).encode()
        )
        __password: str = hash_obj.hexdigest()
        __mfa_code: str = getpass.getpass(
            Print.fmt_info(
                "请输入双重验证码(已隐藏)(如未设置请直接回车):", "§6 MFA  "
            )
        )
        Authorization: str = requests.get(url=constants.FB_APIS[1], timeout=5).text
        repo: requests.Response = requests.post(
            url=constants.FB_APIS[3],
            data=json.dumps(
                {
                    "username": username,
                    "password": __password,
                    "mfa_code": __mfa_code,
                },
                ensure_ascii=False,
            ),
            headers={
                "Content-Type": "application/json",
                "authorization": f"Bearer {Authorization}",
            },
        )
        repo_text: dict = json.loads(repo.text)
        repo_message: str = repo_text["message"]
        repo_success: bool = repo_text["success"]
        if repo.status_code != 200:
            Print.print_war(
                f"请求Api接口失败，将自动使用Token登陆! 状态码:{repo.status_code}，返回值:{repo.text}"
            )
            return
        if not repo_success:
            if "Invalid username, password, or MFA code." in repo_message:
                Print.print_war("登陆失败; 无效的用户名、密码或MFA代码!")
            else:
                Print.print_war("登录失败, 原因: " + repo_message + ", 请重试!")
        else:
            break
    with_perfix: dict = json.loads(
        requests.get(
            url=constants.FB_APIS[2],
            data=json.dumps({"with_prefix": constants.FB_APIS[4]}),
            timeout=5,
            headers={
                "Content-Type": "application/json",
                "authorization": f"Bearer {Authorization}",
            },
        ).text
    )
    fetch_announcements: dict = json.loads(
        requests.get(
            url=constants.FB_APIS[4] + with_perfix["fetch_announcements"],
            timeout=5,
            headers={
                "Content-Type": "application/json",
                "authorization": f"Bearer {Authorization}",
            },
        ).text
    )
    fetch_profile: dict = json.loads(
        requests.get(
            url=constants.FB_APIS[4] + with_perfix["fetch_profile"],
            timeout=5,
            headers={
                "Content-Type": "application/json",
                "authorization": f"Bearer {Authorization}",
            },
        ).text
    )
    get_helper_status: dict = json.loads(
        requests.get(
            url=constants.FB_APIS[4] + with_perfix["get_helper_status"],
            timeout=5,
            headers={
                "Content-Type": "application/json",
                "authorization": f"Bearer {Authorization}",
            },
        ).text
    )
    UserInfo: dict = {
        "isadmin": repo_text["isadmin"],
        "blc": fetch_profile["blc"],
            "cn_username": fetch_profile["cn_username"],
            "phoenix_otp": fetch_profile["phoenix_otp"],
            "points": fetch_profile["points"],
            "slots": fetch_profile["slots"],
            "get_helper_status": get_helper_status,
        }
    token = requests.get(
        url=constants.FB_APIS[4] + with_perfix["get_phoenix_token"],
        data=json.dumps({"secret": f"{Authorization}"}),
        timeout=5,
        headers={
            "Content-Type": "application/json",
            "authorization": f"Bearer {Authorization}",
        },
    ).text
    with open("fbtoken", "w", encoding="utf-8") as f:
        f.write(token)


from .basic_mods import (
    os,
    requests,
    re,
    tarfile,
    gzip,
    json,
    importlib,
    threading,
    time,
)
from .urlmethod import download_file_singlethreaded
from typing import Dict
from .color_print import Print
from glob import glob


class GameTextsLoader:
    def __init__(self) -> None:
        self.base_path = os.path.join(os.getcwd(), "插件数据文件", "game_texts")
        self.check_initial_run()
        self.start_auto_update_thread()
        self.game_texts_data: Dict[str, str] = self.load_data()

    @staticmethod
    def get_latest_version() -> str:
        return re.match(
            r"(\d+\.\d+\.\d+)",
            requests.get(
                "https://api.github.com/repos/ToolDelta/GameText/releases/latest"
            ).json()["tag_name"],
        ).group()

    def check_initial_run(self) -> None:
        version_file_path: str = os.path.join(self.base_path, "version")
        if not os.path.exists(version_file_path):
            latest_version: str = self.get_latest_version()
            with open(version_file_path, "w", encoding="utf-8") as f:
                f.write(latest_version)
            self.download_and_extract(latest_version)

    def start_auto_update_thread(self) -> None:
        threading.Timer(24 * 60 * 60, self.auto_update).start()

    def auto_update(self) -> None:
        version_file_path: str = os.path.join(self.base_path, "version")
        with open(version_file_path, "r", encoding="utf-8") as f:
            version: str = f.read()
        latest_version: str = self.get_latest_version()
        if version != latest_version:
            self.download_and_extract(latest_version)
        threading.Timer(24 * 60 * 60, self.auto_update).start()

    def download_and_extract(self, version):
        packets_url: str = (
            f"https://hub.gitmirror.com/?q=https://github.com/ToolDelta/GameText/releases/download/{version}/ToolDelta_Game_Texts.tar.gz"
        )
        archive_path = os.path.join(self.base_path, "ToolDelta_Game_Texts.tar.gz")
        download_file_singlethreaded(packets_url, archive_path)
        self.extract_data_archive(archive_path)

    def load_data(self) -> Dict[str, str]:
        try:
            all_values: Dict[str, str] = {}
            for file_path in glob(
                os.path.join(self.base_path, "src", "**", "*.py"), recursive=True
            ):
                module_name: str = os.path.basename(file_path).replace(".py", "")
                spec: importlib.util.spec_from_file_location = (
                    importlib.util.spec_from_file_location(module_name, file_path)
                )
                module: importlib.util.module_from_spec = (
                    importlib.util.module_from_spec(spec)
                )
                spec.loader.exec_module(module)
                for var_name in dir(module):
                    if not var_name.startswith("__"):
                        var_value = getattr(module, var_name)
                        if isinstance(var_value, dict):
                            all_values.update(var_value)
            return all_values
        except Exception as err:
            return {}

    def extract_data_archive(self, zip_path: str) -> bool:
        try:
            with gzip.open(zip_path, "rb") as f_in, tarfile.open(
                fileobj=f_in, mode="r"
            ) as tar:
                tar.extractall(self.base_path)
                with open(
                    os.path.join(self.base_path, "src_tree.json"), "w", encoding="utf-8"
                ) as f:
                    json.dump(tar.getnames(), f)
                return True
        except Exception as err:
            Print.print_war(f"Error extracting data archive: {err}")
            return False



from hmac import new
import platform
import os
from poplib import CR
import shlex
import subprocess
import uuid
import time
import json
import ssl
import requests
import ujson
import random
from typing import Callable, Optional

from .color_print import Print
from .urlmethod import download_file_singlethreaded, get_free_port
from .builtins import Builtins
from .basic_mods import socketio
from .packets import Packet_CommandOutput, PacketIDS
from .sys_args import sys_args_to_dict

from .fb_conn import fbconn
import threading

import tooldelta


class SysStatus:
    LOADING = 100
    LAUNCHING = 101
    RUNNING = 102
    NORMAL_EXIT = 103
    FB_LAUNCH_EXC = 104
    CRASHED_EXIT = 105
    NEED_RESTART = 106
    launch_type = "None"


class StandardFrame:
    # 提供了标准的启动器框架, 作为 ToolDelta 和游戏交互的接口
    launch_type = "Original"

    def __init__(self, serverNumber, password, fbToken, auth_server_url):
        self.serverNumber = serverNumber
        self.serverPassword = password
        self.fbToken = fbToken
        self.auth_server = auth_server_url
        self.status = SysStatus.LOADING
        self.system_type = platform.uname().system
        self.inject_events = []
        self.packet_handler: Callable | None = lambda pckType, pck: None
        self.need_listen_packets = {9, 63, 79}
        self._launcher_listener = None
        self.exit_event = threading.Event()

    def add_listen_packets(self, *pcks: int):
        for i in pcks:
            self.need_listen_packets.add(i)

    @staticmethod
    def launch():
        raise Exception("Cannot launch this launcher")

    def listen_launched(self, cb):
        self._launcher_listener = cb

    def close_fb(self):
        raise NotImplementedError()

    @staticmethod
    def get_players_and_uuids():
        return None

    @staticmethod
    def get_bot_name():
        return None

    def update_status(self, new_status):
        self.status = new_status
        if new_status == SysStatus.NORMAL_EXIT:
            tooldelta.safe_jump(out_task=True)
            self.exit_event.set()  # 设置事件，触发等待结束
        if new_status == SysStatus.CRASHED_EXIT:
            tooldelta.safe_jump(out_task=False)
            self.exit_event.set()

    get_all_players = None
    sendcmd: Callable[[str, bool, int | float], None | Packet_CommandOutput]
    sendwscmd: Callable[[str, bool, int | float], Packet_CommandOutput]
    sendwocmd: Callable[[str], None]
    sendfbcmd: Callable[[str], None | AttributeError]
    sendPacket: Callable[[int, str], None]
    sendPacketJson: Callable[[int, str], None]
    is_op: Callable[[str], None | bool] | None


class FrameFBConn(StandardFrame):
    # 使用原生 FastBuilder External 连接
    cmds_reqs = []
    cmds_resp = {}

    def __init__(self, serverNumber, password, fbToken, auth_server):
        super().__init__(serverNumber, password, fbToken, auth_server)
        self.injected = False
        self.downloadMissingFiles()
        self.init_all_functions()

    def launch(self):
        try:
            free_port = get_free_port(10000)
            self.runFB(port=free_port)
            self.run_conn(port=free_port)
            Builtins.createThread(self.output_fb_msgs_thread)
            self.process_game_packets()
        except Exception as err:
            return err

    def runFB(self, ip="127.0.0.1", port=8080):
        if self.system_type == "Linux":
            os.system("/usr/bin/chmod +x ./phoenixbuilder")
            con_cmd = rf"./phoenixbuilder -A {self.auth_server} -t fbtoken --no-readline --no-update-check --listen-external {ip}:{port} -c {self.serverNumber} {f'-p {self.serverPassword}' if self.serverPassword else ''}"

        # windows updated "./PRGM" command.
        if self.system_type == "Windows":
            con_cmd = rf".\phoenixbuilder.exe -A {self.auth_server} -t fbtoken --no-readline --no-update-check --listen-external {ip}:{port} -c {self.serverNumber} {f'-p {self.serverPassword}' if self.serverPassword else ''}"
        self.fb_pipe = subprocess.Popen(
            con_cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

    def run_conn(self, ip="127.0.0.1", port=8080, timeout=None):
        connect_fb_start_time = time.time()
        max_con_time = timeout or 10
        while 1:
            try:
                self.con = fbconn.ConnectFB(f"{ip}:{port}")
                Print.print_suc("§a成功连接上FastBuilder.")
                return 1
            except:
                if time.time() - connect_fb_start_time > max_con_time:
                    Print.print_err(f"§4{max_con_time}秒内未连接上FB，已退出")
                    self.close_fb()
                    raise SystemExit
                if self.status == SysStatus.FB_LAUNCH_EXC:
                    Print.print_err("§4连接FB时出现问题，已退出")
                    self.close_fb()
                    raise SystemExit

    def output_fb_msgs_thread(self):
        while 1:
            tmp: str = self.fb_pipe.stdout.readline().decode("utf-8").strip("\n")
            if not tmp:
                continue
            if " 简体中文" in tmp:
                # seems will be unable forever because it's no longer supported.
                try:
                    self.fb_pipe.stdin.write(f"{tmp[1]}\n".encode("utf-8"))
                    self.fb_pipe.stdin.flush()
                    Print.print_inf(f"语言已自动选择为简体中文： [{tmp[1]}]")
                except IndexError:
                    Print.print_war("未能自动选择为简体中文")
            elif "ERROR" in tmp:
                if "Server not found" in tmp:
                    Print.print_err(
                        f"§c租赁服号: {self.serverNumber} 未找到, 有可能是租赁服关闭中, 或是设置了等级或密码"
                    )
                    self.update_status(SysStatus.CRASHED_EXIT)

                elif "Unauthorized rental server number" in tmp:
                    Print.print_err(
                        f"§c租赁服号: {self.serverNumber} ，你还没有该服务器号的卡槽， 请前往用户中心购买"
                    )
                    self.update_status(SysStatus.CRASHED_EXIT)
                elif "Failed to contact with API" in tmp:
                    Print.print_err(
                        "§c无法连接到验证服务器, 可能是FB服务器崩溃, 或者是你的IP处于黑名单中"
                    )
                    try:
                        Print.print_war("尝试连接到 FastBuilder 验证服务器")
                        requests.get("http://user.fastbuilder.pro", timeout=10)
                        Print.print_err(
                            "??? 未知情况， 有可能只是验证服务器崩溃， 用户中心并没有崩溃"
                        )
                    except:
                        Print.print_err(
                            "§cFastBuilder服务器无法访问， 请等待修复(加入FastBuilder频道查看详情)"
                        )
                    self.update_status(SysStatus.CRASHED_EXIT)
                elif "Invalid token" in tmp:
                    Print.print_err("§cFastBuilder Token 无法使用， 请重新下载")
                    self.update_status(SysStatus.CRASHED_EXIT)
                elif "netease.report.kick.hint" in tmp:
                    Print.print_err(
                        "§c无法连接到网易租赁服 -> 网易土豆的常见问题，检查你的租赁服状态（等级、是否开启、密码）并重试, 也可能是你的网络问题"
                    )
                    self.update_status(SysStatus.CRASHED_EXIT)
                elif "Press ENTER to exit." in tmp:
                    Print.print_err("§c程序退出")
                    self.update_status(SysStatus.CRASHED_EXIT)
                else:
                    Print.print_with_info(tmp, "§b  FB  §r")

            elif "Transfer: accept new connection @ " in tmp:
                Print.print_with_info(
                    "FastBuilder 监听端口已开放: " + tmp.split()[-1], "§b  FB  "
                )
            elif "Successfully created minecraft dialer." in tmp:
                Print.print_with_info("§e成功创建于 Minecraft 的链接", "§b  FB  §r")
            elif tmp.startswith("panic"):
                Print.print_err(f"FastBuilder 出现问题: {tmp}")
            else:
                Print.print_with_info(tmp, "§b  FB  §r")

    def close_fb(self):
        try:
            self.fb_pipe.stdin.write("exit\n".encode("utf-8"))
            self.fb_pipe.stdin.flush()
        except:
            pass
        try:
            self.fb_pipe.kill()
        except:
            pass
        Print.print_suc("成功关闭FB进程")

    def process_game_packets(self):
        try:
            for packet_bytes in fbconn.RecvGamePacket(self.con):
                packet_type = packet_bytes[0]
                if packet_type not in self.need_listen_packets:
                    continue
                packet_mapping = ujson.loads(
                    fbconn.GamePacketBytesAsIsJsonStr(packet_bytes)
                )
                if packet_type == PacketIDS.CommandOutput:
                    cmd_uuid = packet_mapping["CommandOrigin"]["UUID"].encode()
                    if cmd_uuid in self.cmds_reqs:
                        self.cmds_resp[cmd_uuid] = [time.time(), packet_mapping]
                self.packet_handler(packet_type, packet_mapping)
                if not self.injected and packet_type == PacketIDS.PlayerList:
                    self.injected = True
                    Builtins.createThread(self._launcher_listener)
        except StopIteration:
            pass
        self.update_status(SysStatus.CRASHED_EXIT)

    def downloadMissingFiles(self):
        "获取缺失文件"
        Print.print_with_info("§d将自动检测缺失文件并补全", "§d 加载 ")
        mirror_src = "https://mirror.ghproxy.com/"
        file_get_src = (
            mirror_src
            + "https://raw.githubusercontent.com/ToolDelta/ToolDelta/main/require_files.json"
        )
        try:
            files_to_get = json.loads(requests.get(file_get_src, timeout=30).text)
        except json.JSONDecodeError:
            Print.print_err("自动下载缺失文件失败: 文件源 JSON 不合法")
            return False
        except requests.Timeout:
            Print.print_err("自动下载缺失文件失败: URL 请求出现问题: 请求超时")
            return False
        except Exception as err:
            Print.print_err(f"自动下载缺失文件失败: URL 请求出现问题: {err}")
            return False
        try:
            Print.print_with_info("§d正在检测需要补全的文件", "§d 加载 ")
            mirrs = files_to_get["Mirror"]
            files = files_to_get[self.system_type]
            for fdir, furl in files.items():
                if not os.path.isfile(fdir):
                    Print.print_inf(f"文件: <{fdir}> 缺失, 正在下载..")
                    succ = False
                    for mirr in mirrs:
                        try:
                            download_file_singlethreaded(
                                mirr + "/https://github.com/" + furl, fdir
                            )
                            succ = True
                            break
                        except requests.exceptions.RequestException:
                            Print.print_war("镜像源故障, 正在切换")
                    if not succ:
                        Print.print_err("镜像源全不可用..")
                        return False
                    Print.print_inf(f"文件: <{fdir}> 下载完成        ")
        except requests.Timeout:
            Print.print_err("自动检测文件并补全时出现错误: 超时, 自动跳过")
        except Exception as err:
            Print.print_err(f"自动检测文件并补全时出现错误: {err}")
            return False
        return True

    def init_all_functions(self):
        def sendcmd(cmd: str, waitForResp: bool = False, timeout: int | float = 30):
            uuid = fbconn.SendMCCommand(self.con, cmd)
            if waitForResp:
                self.cmds_reqs.append(uuid)
                waitStartTime = time.time()
                while 1:
                    res = self.cmds_resp.get(uuid)
                    if res is not None:
                        self.cmds_reqs.remove(uuid)
                        del self.cmds_resp[uuid]
                        return Packet_CommandOutput(res[1])
                    if time.time() - waitStartTime > timeout:
                        self.cmds_reqs.remove(uuid)
                        Print.print_war(f'sendcmd "{cmd}" 超时, 尝试 sendwscmd')
                        return self.sendwscmd(cmd, True, timeout)

        def sendwscmd(cmd: str, waitForResp: bool = False, timeout: int = 30):
            uuid = fbconn.SendWSCommand(self.con, cmd)
            if waitForResp:
                self.cmds_reqs.append(uuid)
                waitStartTime = time.time()
                while 1:
                    res = self.cmds_resp.get(uuid)
                    if res is not None:
                        self.cmds_reqs.remove(uuid)
                        del self.cmds_resp[uuid]
                        return Packet_CommandOutput(res[1])
                    if time.time() - waitStartTime > timeout:
                        self.cmds_reqs.remove(uuid)
                        raise TimeoutError("指令超时")
            else:
                return uuid

        self.sendcmd = sendcmd
        self.sendwscmd = sendwscmd
        self.sendwocmd = staticmethod(
            lambda cmd: fbconn.SendNoResponseCommand(self.con, cmd)
        )
        self.sendPacket = self.sendPacketJson = staticmethod(
            lambda pckID, pck: fbconn.SendGamePacketBytes(
                self.con,
                fbconn.JsonStrAsIsGamePacketBytes(
                    pckID, ujson.dumps(pck, ensure_ascii=False)
                ),
            )
        )
        self.sendfbcmd = staticmethod(lambda cmd: fbconn.SendFBCommand(self.con, cmd))
        self.is_op = None


class FrameNeOmg(StandardFrame):
    # 使用 NeOmega 框架连接到游戏
    launch_type = "NeOmega"

    def __init__(self, serverNumber, password, fbToken, auth_server):
        super().__init__(serverNumber, password, fbToken, auth_server)
        self.status = None
        self.launch_event = threading.Event()
        self.injected = False
        self.omega = None
        self.TDC = None
        self.set_tooldelta_cli()
        self.download_libs()
        self.init_all_functions()
        self.status = SysStatus.LOADING
        self.secret_exit_key = ""

    def set_omega(self, openat_port):
        from .neo_libs import neo_conn

        retries = 0
        while retries <= 10:
            try:
                self.omega = neo_conn.ThreadOmega(
                    connect_type=neo_conn.ConnectType.Remote,
                    address=f"tcp://localhost:{openat_port}",
                    accountOption=neo_conn.AccountOptions(
                        AuthServer=self.auth_server,
                        UserToken=self.fbToken,
                        ServerCode=self.serverNumber,
                        ServerPassword=str(self.serverPassword),
                    ),
                )
                retries = 0
                break
            except Exception as err:
                time.sleep(5)
                retries += 1
                Print.print_war(f"OMEGA 连接失败, 重连: 第 {retries} 次: {err}")
                if retries > 5:
                    Print.print_err("最大重试次数已达到")
                    raise SystemExit

    def set_tooldelta_cli(self):
        if type(self) == FrameNeOmg:
            self.TDC = ToolDeltaCli()

    def start_neomega_proc(self):
        free_port = get_free_port(24016)
        access_point_file = (
            f"neomega_{platform.uname().system.lower()}_access_point_{self.sys_machine}"
        )
        if "TERMUX_VERSION" in os.environ:
            access_point_file = f"neomega_android_access_point_{self.sys_machine}"
        if platform.system() == "Windows":
            access_point_file += ".exe"
        py_file_path = os.path.join(
            os.getcwd(), "tooldelta", "neo_libs", access_point_file
        )
        if platform.uname().system.lower() == "linux":
            os.system("chmod +x " + shlex.quote(py_file_path))
        # 只需要+x即可
        Print.print_inf(f"DEBUG: 将使用端口 {free_port}")
        self.neomg_proc = subprocess.Popen(
            [
                py_file_path,
                "-server",
                self.serverNumber,
                "-T",
                self.fbToken,
                "-access-point-addr",
                f"tcp://localhost:{free_port}",
                "-server-password",
                str(self.serverPassword),
                "-auth-server",
                self.auth_server,
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        return free_port

    def msg_show(self):
        def _msg_show_thread():
            while True:
                msg_orig = self.neomg_proc.stdout.readline().decode("utf-8").strip("\n")
                if msg_orig in ("", "SIGNAL: exit"):
                    with Print.lock:
                        Print.print_with_info("ToolDelta: NEOMG 进程已结束", "§b NOMG ")
                    self.update_status(SysStatus.NORMAL_EXIT)
                    return
                if "[neOmega 接入点]: 就绪" in msg_orig:
                    self.launch_event.set()
                elif f"STATUS CODE: {self.secret_exit_key}" in msg_orig:
                    with Print.lock:
                        Print.print_with_info("§a机器人已退出", "§b NOMG ")
                    continue
                with Print.lock:
                    Print.print_with_info(msg_orig, "§b NOMG ")

        Builtins.createThread(_msg_show_thread, usage="显示来自NeOmega的信息")

    def make_secret_key(self):
        self.secret_exit_key = hex(random.randint(10000, 99999))

    def launch(self):
        self.status = SysStatus.LAUNCHING
        openat_port = self.start_neomega_proc()
        self.msg_show()
        self.launch_event.wait()
        self.make_secret_key()
        self.set_omega(openat_port)
        self.update_status(SysStatus.RUNNING)
        Print.print_suc("已开启接入点进程")
        pcks = [
            self.omega.get_packet_id_to_name_mapping(i)
            for i in self.need_listen_packets
        ]
        self.omega.listen_packets(pcks, self.packet_handler_parent)
        self._launcher_listener()
        Print.print_suc("接入点已就绪!")
        self.exit_event.wait()  # 等待事件的触发
        if self.status == SysStatus.NORMAL_EXIT:
            return SystemExit("正常退出.")
        if self.status == SysStatus.CRASHED_EXIT:
            return Exception("NeOmega 已崩溃")
        return SystemError("未知的退出状态")

    def download_libs(self):
        if self.TDC.SocketIO.connected:
            try:
                res = self.TDC.get_depends_table_data()
                use_mirror = res["mirror_data"]["Mirror"][0]
            except Exception as err:
                Print.print_err(f"获取依赖库表出现问题: {err}")
                self.update_status(SysStatus.CRASHED_EXIT)
                return
            self.sys_machine = platform.machine().lower()
            if self.sys_machine == "x86_64":
                self.sys_machine = "amd64"
            elif self.sys_machine == "aarch64":
                self.sys_machine = "arm64"
            if "TERMUX_VERSION" in os.environ:
                sys_info_fmt = f"Android:{self.sys_machine.lower()}"
            else:
                sys_info_fmt = f"{platform.uname().system}:{self.sys_machine.lower()}"
            source_dict = res["mirror_data"][sys_info_fmt]
            commit_file_path = res["commit_file_path"]
            commit_url = res["commit"]
            commit_remote = res["commit"]
            commit_local = ""
            commit_file_path = os.path.join(os.getcwd(), commit_file_path)
            replace_file = False
            if os.path.isfile(commit_file_path):
                with open(commit_file_path, "r", encoding="utf-8") as f:
                    commit_local = f.read()
                if commit_local != commit_remote:
                    Print.print_war("依赖库版本过期, 将重新下载")
                    replace_file = True
            else:
                replace_file = True
            for k, v in source_dict.items():
                pathdir = os.path.join(os.getcwd(), k)
                url = use_mirror + "/https://raw.githubusercontent.com/" + v
                if not os.path.isfile(pathdir) or replace_file:
                    Print.print_with_info(f"正在下载依赖库 {pathdir} ...", "§a 下载 §r")
                    try:
                        download_file_singlethreaded(url, pathdir)
                    except Exception as err:
                        Print.print_err(f"下载依赖库出现问题: {err}")
                        self.update_status(SysStatus.CRASHED_EXIT)
                        return
            if replace_file:
                # 写入commit_remote，文字写入
                with open(commit_file_path, "w", encoding="utf-8") as f:
                    f.write(commit_remote)
                Print.print_suc("已完成依赖更新！")

        elif not self.TDC.SocketIO.connected:
            try:
                res = json.loads(
                    requests.get(
                        "https://mirror.ghproxy.com/https://raw.githubusercontent.com/ToolDelta/ToolDelta/main/require_files.json"
                    ).text
                )
                use_mirror = res["Mirror"][0]
            except Exception as err:
                Print.print_err(f"获取依赖库表出现问题: {err}")
                self.update_status(SysStatus.CRASHED_EXIT)
                return
            self.sys_machine = platform.machine().lower()
            if self.sys_machine == "x86_64":
                self.sys_machine = "amd64"
            elif self.sys_machine == "aarch64":
                self.sys_machine = "arm64"
            if "TERMUX_VERSION" in os.environ:
                sys_info_fmt = f"Android:{self.sys_machine.lower()}"
            else:
                sys_info_fmt = f"{platform.uname().system}:{self.sys_machine.lower()}"
            source_dict = res[sys_info_fmt]
            commit_file_path, commit_url = list(res["Commit"].items())[0]
            commit_remote = requests.get(
                use_mirror + "/raw.githubusercontent.com/" + commit_url
            ).text
            commit_local = ""
            commit_file_path = os.path.join(os.getcwd(), commit_file_path)
            replace_file = False
            if os.path.isfile(commit_file_path):
                with open(commit_file_path, "r", encoding="utf-8") as f:
                    commit_local = f.read()
                if commit_local != commit_remote:
                    Print.print_war("依赖库版本过期, 将重新下载")
                    replace_file = True
            else:
                replace_file = True
            for k, v in source_dict.items():
                pathdir = os.path.join(os.getcwd(), k)
                url = use_mirror + "/https://raw.githubusercontent.com/" + v
                if not os.path.isfile(pathdir) or replace_file:
                    Print.print_with_info(f"正在下载依赖库 {pathdir} ...", "§a 下载 §r")
                    try:
                        download_file_singlethreaded(url, pathdir)
                    except Exception as err:
                        Print.print_err(f"下载依赖库出现问题: {err}")
                        self.update_status(SysStatus.CRASHED_EXIT)
                        return
            if replace_file:
                # 写入commit_remote，文字写入
                with open(commit_file_path, "w", encoding="utf-8") as f:
                    f.write(commit_remote)
                Print.print_suc("已完成依赖更新！")

    def get_players_and_uuids(self):
        players_uuid = {}
        for i in self.omega.get_all_online_players():
            players_uuid[i.name] = i.uuid
        return players_uuid

    def get_bot_name(self):
        return self.omega.get_bot_name()

    def packet_handler_parent(self, pkt_type, pkt):
        pkt_type = self.omega.get_packet_name_to_id_mapping(pkt_type)
        self.packet_handler(pkt_type, pkt)

    def init_all_functions(self):
        def sendcmd(cmd: str, waitForResp: bool = False, timeout: int | float = 30):
            if waitForResp:
                res = self.omega.send_player_command_need_response(cmd, timeout)
                if res is None:
                    raise TimeoutError("指令超时")
                return res
            self.omega.send_player_command_omit_response(cmd)
            return

        def sendwscmd(cmd: str, waitForResp: bool = False, timeout: int = 30):
            if waitForResp:
                res = self.omega.send_websocket_command_need_response(cmd, timeout)
                if res is None:
                    raise TimeoutError("指令超时")
                return res
            self.omega.send_websocket_command_omit_response(cmd)
            return

        def sendwocmd(cmd: str):
            self.omega.send_settings_command(cmd)

        def sendPacket(pktID: int, pkt: str):
            self.omega.send_game_packet_in_json_as_is(pktID, pkt)

        def sendfbcmd(_):
            raise AttributeError("NeOmg模式无法发送FBCommand")

        def is_op(player: str):
            return self.omega.get_player_by_name(player).command_permission_level > 2

        self.sendcmd = sendcmd
        self.sendwscmd = sendwscmd
        self.sendwocmd = sendwocmd
        self.sendPacket = self.sendPacketJson = sendPacket
        self.sendfbcmd = sendfbcmd
        self.is_op = is_op


class FrameNeOmgRemote(FrameNeOmg):
    def launch(self):
        try:
            openat_port = int(sys_args_to_dict().get("access-point-port", "24020"))
            if openat_port not in range(65536):
                raise AssertionError
        except (ValueError, AssertionError):
            Print.print_err("启动参数 -access-point-port 错误: 不是1~65535的整数")
        if openat_port == 0:
            Print.print_war(
                "未用启动参数指定链接neOmega接入点开放端口, 尝试使用默认端口 24015"
            )
            Print.print_inf("可使用启动参数 -access-point-port 端口 以指定接入点端口.")
            openat_port = 24015
            return SystemExit
        Print.print_inf(f"将从端口 {openat_port} 连接至接入点(等待接入中).")
        self.set_omega(openat_port)
        Print.print_suc("已连接上接入点进程.")
        pcks = [
            self.omega.get_packet_id_to_name_mapping(i)
            for i in self.need_listen_packets
        ]
        self.omega.listen_packets(pcks, self.packet_handler_parent)
        self._launcher_listener()
        Print.print_suc("接入点已就绪")
        self.exit_event.wait()
        self.update_status(SysStatus.NORMAL_EXIT)
        if self.status == SysStatus.NORMAL_EXIT:
            return SystemExit("正常退出.")
        if self.status == SysStatus.CRASHED_EXIT:
            return Exception("接入点已崩溃")
        return SystemError("未知的退出状态")

    def download_libs(self):
        Print.print_inf("以 Remote 启动, 将不会检查库完整性")

class ToolDeltaCli(object):
    def __init__(self, address: dict = {"host": "tdaus.tooldelta.fit", "port": 0}) -> None:
    # def __init__(self, address: dict = {"host": "127.0.0.1", "port": 9002}) -> None:
        self.NoPort:bool = address.get("port", 0) == 0
        self.S_ADDRESSS:dict = address
        self.protocol:str = "http"
        self.url:str = f'{self.protocol}://{self.S_ADDRESSS["host"]}:{self.S_ADDRESSS["port"]}' if self.NoPort ==False else f'{self.protocol}://{self.S_ADDRESSS["host"]}'
        self.SocketIO:socketio.Client = socketio.Client()
        self.data_received_event:threading.Event = threading.Event()
        self.connected_to_server:bool = True
        threading.Thread(target=self.conn_aus, name="SocketIO_Conn").start()
        while not self.SocketIO.connected and self.connected_to_server:time.sleep(0.1)

    def init_auth_v(self) -> None:
        self.feature_code:str = str(uuid.uuid5(uuid.NAMESPACE_DNS, str(time.perf_counter())))
        self.token_ec:tuple= (self.feature_code, self.get_new_token())

    def get_new_token(self) -> None:
        try:
            response = requests.post(url=f'{self.url}/api/signin', data=json.dumps({"feature_code": self.feature_code}), timeout= 5)
            if response.status_code == 200:
                return response.text
        except requests.exceptions.ConnectionError as err:return "null"

    def conn_aus(self) -> None:
        try:
            self.init_auth_v()
            self.SocketIO.connect(self.url, namespaces='/api', headers={'Authorization': f'Bearer {self.token_ec[1]}'})
            Print.print_suc("ToolDelta成功连接到至Api服务器[Socket-IO]!")
            while True:
                if not self.SocketIO.connected:
                    try:
                        self.init_auth_v()
                        self.SocketIO.connect(self.url, namespaces='/api', headers={'Authorization': f'Bearer {self.token_ec[1]}'})
                        Print.print_suc("ToolDelta与Api服务器断开连接,已重新连接成功!")
                    except:
                        Print.print_war("ToolDeltaApi服务器可能存在故障或当前网络环境异常，将停止使用ToolDeltaApi服务器!")
                        break
                time.sleep(10)
        except Exception as err:
            Print.print_war("ToolDelta无法连接至Api服务器,将停止使用其提供的功能!")
            self.connected_to_server = False
            threading.Thread(target=self.reconnect_to_server).start()

    def reconnect_to_server(self, interval=20):
        while not self.connected_to_server:
            try:
                self.init_auth_v()
                self.SocketIO.connect(self.url, namespaces='/api', headers={'Authorization': f'Bearer {self.token_ec[1]}'})
                Print.print_suc("ToolDelta成功重新连接至Api服务器!")
                self.connected_to_server = True
            except Exception as err:
                time.sleep(interval)

    def get_depends_table_data(self) -> dict:
        if self.SocketIO.connected:
            @self.SocketIO.on('depends_table_data', namespace='/api')
            def handle_depends_table_data(data):
                self.data_received_event.set()
                self.depend_table_data = data
            self.SocketIO.emit('get_depends_table', namespace='/api')
            self.data_received_event.wait()  # 等待数据返回
            self.data_received_event.clear()
            return self.depend_table_data
        else:
            Print.print_war("Namespace /api is not connected yet. Please wait for connection.")

    def get_version_updates(self) -> any:
        if self.SocketIO.connected:
            @self.SocketIO.on('version_updates', namespace='/api')
            def handle_version_updates_data(data):
                self.data_received_event.set()
                self.latest_version_data = data

            self.SocketIO.emit('get_version_update', namespace='/api')
            self.data_received_event.wait()
            self.data_received_event.clear()
            return self.latest_version_data["latest_version_str"]


import datetime
import colorama
import threading

colorama.init(autoreset=True)

try:
    from .logger import publicLogger
except ImportError:
    pass

unicode = type("")


def simple_fmt(kw, __sub):
    for k, v in kw.items():
        if k in __sub:
            __sub = __sub.replace(k, str(v))
    return __sub


class _Print:
    INFO_NORMAL = "§f 信息 "
    INFO_WARN = "§6 警告 "
    INFO_ERROR = "§4 报错 "
    INFO_FAIL = "§c 失败 "
    INFO_SUCC = "§a 成功 "
    INFO_LOAD = "§d 加载 "
    STD_COLOR_LIST = [
        ["0", "#000000"],
        ["1", "#0000AA"],
        ["2", "#00AA00"],
        ["3", "#00AAAA"],
        ["4", "#AA0000"],
        ["5", "#AA00AA"],
        ["6", "#FFAA00"],
        ["7", "#AAAAAA"],
        ["8", "#555555"],
        ["9", "#5555FF"],
        ["a", "#55FF55"],
        ["b", "#55FFFF"],
        ["c", "#FF5555"],
        ["d", "#FF55FF"],
        ["e", "#FFFF55"],
        ["f", "#FFFFFF"],
        ["g", "#DDD605"],
        ["r", "/"],
    ]

    def __init__(self):
        self.lock = threading.RLock()

    def __call__(self, text: str):
        with self.lock:
            self.print_inf(text)

    def simple_fmt(self, kw, __sub):
        with self.lock:
            return simple_fmt(kw, __sub)

    def colormode_replace(self, text: str, showmode=0):
        with self.lock:
            # 1 = bg_color
            text = self._strike(text)
            return (
                simple_fmt(
                    {
                        "§1": f"\033[{showmode};37;34m",
                        "§2": f"\033[{showmode};37;32m",
                        "§3": f"\033[{showmode};37;36m",
                        "§4": f"\033[{showmode};37;31m",
                        "§5": f"\033[{showmode};37;35m",
                        "§6": f"\033[{showmode};37;33m",
                        "§7": f"\033[{showmode};37;90m",
                        "§8": f"\033[{showmode};37;2m",
                        "§9": f"\033[{showmode};37;94m",
                        "§a": f"\033[{showmode};37;92m",
                        "§b": f"\033[{showmode};37;96m",
                        "§c": f"\033[{showmode};37;91m",
                        "§d": f"\033[{showmode};37;95m",
                        "§e": f"\033[{showmode};37;93m",
                        "§f": f"\033[{showmode};37;1m",
                        "§r": "\033[0m",
                        "§u": "\033[4m",
                        "§l": "\033[1m",
                    },
                    text,
                )
                + "\033[0m"
            )


    @staticmethod
    def align(text: str, length=15):
        l = len(text)
        for char in text:
            if not char.isascii():
                l += 1
        return text + " " * (length - l)

    @staticmethod
    def _strike(text: str):
        text_ok = ""
        strikeMode = False
        i = 0
        while i < len(text):
            char = text[i]
            try:
                if char == "§":
                    if text[i + 1] == "s":
                        strikeMode = True
                        i += 2
                        continue
                    if text[i + 1] == "r":
                        strikeMode = False
            except IndexError:
                pass
            if strikeMode:
                text_ok += "\u0336" + char
            else:
                text_ok += char
            i += 1
        return text_ok


    def print_with_info(
        self, text: str, info: str = INFO_NORMAL, need_log=True, **print_kwargs
    ):
        with self.lock:
            if need_log:
                self.c_log(info, text)
            setNextColor = "§r"
            if "\n" in text:
                output_txts = []
                for text_line in str(text).split("\n"):
                    if "§" in text_line:
                        try:
                            n = text_line.rfind("§")
                            _setNextCol = text_line[n : n + 2]
                            if setNextColor == -1:
                                raise AssertionError
                            setNextColor = _setNextCol
                        except:
                            pass
                    output_txts.append(
                        datetime.datetime.now().strftime("[%H:%M] ")
                        + self.colormode_replace(info, 7)
                        + " "
                        + self.colormode_replace(setNextColor + text_line)
                    )
                print("\n".join(output_txts), **print_kwargs)
            else:
                print(
                    datetime.datetime.now().strftime("[%H:%M] ")
                    + self.colormode_replace(info, 7)
                    + " "
                    + self.colormode_replace(text),
                    **print_kwargs,
                )


    def clean_print(self, text: str, **print_kwargs):
        with self.lock:
            print(self.colormode_replace(text), **print_kwargs)

    def clean_fmt(self, text: str):
        with self.lock:
            return self.colormode_replace(text)

    def print_err(self, text: str, **print_kwargs):
        with self.lock:
            self.print_with_info(f"§c{text}", self.INFO_ERROR, **print_kwargs)

    def print_inf(self, text: str, **print_kwargs):
        with self.lock:
            self.print_with_info(f"{text}", self.INFO_NORMAL, **print_kwargs)

    def print_suc(self, text: str, **print_kwargs):
        with self.lock:
            self.print_with_info(f"§a{text}", self.INFO_SUCC, **print_kwargs)

    def print_war(self, text: str, **print_kwargs):
        with self.lock:
            self.print_with_info(f"§6{text}", self.INFO_WARN, **print_kwargs)

    def print_load(self, text: str, **print_kwargs):
        with self.lock:
            self.print_with_info(f"§d{text}", self.INFO_LOAD, **print_kwargs)

    def fmt_info(self, text: str, info: str = "§f 信息 "):
        with self.lock:
            setNextColor = "§r"
            if "\n" in text:
                output_txts = []
                for text_line in str(text).split("\n"):
                    if "§" in text_line:
                        try:
                            n = text_line.rfind("§")
                            _setNextCol = text_line[n : n + 2]
                            if setNextColor == -1:
                                raise AssertionError
                            setNextColor = _setNextCol
                        except:
                            pass
                    output_txts.append(
                        datetime.datetime.now().strftime("[%H:%M] ")
                        + self.colormode_replace(info, 7)
                        + " "
                        + self.colormode_replace(setNextColor + text_line)
                    )
                return "\n".join(output_txts)
            return (
                datetime.datetime.now().strftime("[%H:%M] ")
                + self.colormode_replace(info, 7)
                + " "
                + self.colormode_replace(text)
            )

    def c_log(self, inf: str, msg: str):
        with self.lock:
            for _g, _s in [
                ("§6 警告 ", "WARN"),
                ("§a 成功 ", "INFO"),
                ("§f 信息 ", "INFO"),
                ("§c 失败 ", "FAIL"),
                ("§4 报错 ", "ERROR"),
            ]:
                if inf == _g:
                    inf = _s
                    break
            for col, _ in self.STD_COLOR_LIST:
                col = "§" + col
                msg = msg.replace(col, "")
            for col, _ in self.STD_COLOR_LIST:
                col = "§" + col
                inf = inf.replace(col, "")
            inf = inf.replace(" ", "")
            publicLogger.log_in(msg, inf)

Print = _Print()
